#+xcvb (module (:build-depends-on () :depends-on ("/asdf")))
#| ;;; cl-launch 3.000 lisp header
|# ;;;; Silence our lisp implementation for quiet batch use...

#| We'd like to evaluate as little as possible of the code without compilation.
 This poses a typical bootstrapping problem: the more sophistication we want
 to distinguish what to put where in what dynamic environment, the more code
 we have to evaluate before we may actually load compiled files. And, then,
 it is a waste of time to try to compile said code into a file. Moving things
 to the shell can only help so much, and reduces flexibility. Our best bet is
 to tell sbcl or cmucl to not try to optimize too hard.
|#
(eval-when (:compile-toplevel :load-toplevel :execute)
;; #+gcl (trace compile-file load)
(declaim (optimize (speed 2) (safety 3) #-gcl (debug 2) ; (compilation-speed 2)
          #+sbcl (sb-ext:inhibit-warnings 3)
          #+cmu (ext:inhibit-warnings 3)))
#-(or allegro clisp clozure cmu ecl gcl lispworks sbcl)
(error "CL-Launch doesn't support ~A." (lisp-implementation-type))
#+gcl ;;; If using GCL, do some safety checks
(unless (member :ansi-cl *features*)
  (format *error-output* "CL-Launch only supports GCL in ANSI mode.~%")
  (lisp:quit))
#+gcl
(when (or (< system::*gcl-major-version* 2)
          (and (= system::*gcl-major-version* 2)
               (< system::*gcl-minor-version* 7)))
  (pushnew :gcl-pre2.7 *features*))
(setf *print-readably* nil ; allegro 5.0 notably will bork without this
      *print-level* nil
      *load-verbose* nil *compile-verbose* nil *compile-print* nil)
#+(and allegro (version>= 8 0)) (setf excl:*warn-on-nested-reader-conditionals* nil)
#+allegro (setf *error-output* excl::*stderr*)
#+sbcl (proclaim '(sb-ext:muffle-conditions sb-ext:compiler-note))
#+cmu (setf ext:*gc-verbose* nil)
#+clisp (setf custom:*source-file-types* nil custom:*compiled-file-types* nil)
#+gcl (setf compiler::*compiler-default-type* (pathname "")
            compiler::*lsp-ext* "")
#+ecl (require 'cmp)

;;;; Ensure package hygiene
(unless (find-package :cl-launch)
  (make-package :cl-launch
                :use '(#-gcl-pre2.7 :common-lisp #+gcl-pre2.7 :lisp)))
(in-package :cl-launch))
(defmacro dbg (tag &rest exprs)
  "simple debug statement macro:
outputs a tag plus a list of source expressions and their resulting values, returns the last values"
  (let ((res (gensym))(f (gensym)))
  `(let ((,res))
    (flet ((,f (fmt &rest args) (apply #'format *trace-output* fmt args)))
      (,f "~&~A~%" ,tag)
      ,@(mapcan
         #'(lambda (x)
            `((,f "~&  ~S => " ',x)
              (,f "~{~S~^ ~}~%" (setf ,res (multiple-value-list ,x)))))
         exprs)
      (apply 'values ,res)))))
(eval-when (:compile-toplevel :load-toplevel :execute)
(map () #'export
     '(*arguments* getenv quit compile-and-load-file load-systems))
;;; define getenv and quit in ways that minimize package conflicts
;;; (use-package :cl-launch) while in cl-user.
(defun getenv (x)
  #-cmu
  (#+clozure ccl::getenv
   #+allegro sys:getenv
   #+gcl system:getenv
   #+clisp ext:getenv
   #+ecl si:getenv
   #+sbcl sb-ext:posix-getenv
   #+lispworks lispworks:environment-variable x)
  #+cmu (cdr (assoc (intern x :keyword) ext:*environment-list*)))
(defun quit (&optional (code 0) (finish-output t))
  (when finish-output ;; essential, for ClozureCL, and for standard compliance.
    (finish-outputs))
  #+cmu (unix:unix-exit code)
  #+clisp (ext:quit code)
  #+sbcl (sb-unix:unix-exit code)
  #+clozure (ccl:quit code)
  #+gcl (lisp:quit code)
  #+allegro (excl:exit code :quiet t)
  #+ecl (si:quit code)
  #+lispworks (lispworks:quit :status code :confirm nil :return nil :ignore-errors-p t))
(defun compile-file-pathname* (x &rest keys)
  (apply 'compile-file-pathname x #-gcl-pre2.7 keys #+gcl-pre2.7 nil))
#+gcl-pre2.7 (defun ensure-directories-exist (x) "hope for the best" nil)
(defvar *temporary-directory* "/tmp/"
  "The name of the implementation, used to make a directory hierarchy for fasl files")
(defvar *cl-launch-file* nil) ;; name of this very file
(defvar *verbose* nil)
#+(or ecl (and gcl (not gcl-pre2.7)))
(progn
  (defvar *temporary-filenames* nil)
  (defun copy-stream (i o &key (element-type 'character))
    (loop :with size = 8192
          :with buf = (make-array size :element-type element-type)
          :for n = (read-sequence buf i)
          :while (plusp n)
          :do (write-sequence buf o :end n)))
  (defun call-with-new-file (n f)
    (with-open-file (o n :direction :output :if-exists :error :if-does-not-exist :create)
      (funcall f o)))
  (defun dump-stream-to-file (i n)
    (call-with-new-file n (lambda (o) (copy-stream i o))))
  (defun copy-file (src dst)
    (with-open-file (i src :direction :input :if-does-not-exist :error :element-type '(unsigned-byte 8))
      (with-open-file (o dst :direction :output :if-exists :supersede :element-type '(unsigned-byte 8))
        (copy-stream i o :element-type '(unsigned-byte 8)))))
  (defun dump-sexp-to-file (x n)
    (call-with-new-file
     n
     (lambda (o) (write x :stream o :pretty t :readably t))))
  (defvar *temporary-file-prefix*
    (format nil "~Acl-launch-~A-" *temporary-directory* (getenv "CL_LAUNCH_PID")))
  (defun make-temporary-filename (x)
    (concatenate 'string *temporary-file-prefix* x))
  (defun register-temporary-filename (n)
    (push n *temporary-filenames*)
    n)
  (defun temporary-filename (x)
    (register-temporary-filename (make-temporary-filename x)))
  (defun temporary-file-from-foo (dumper arg x)
    (let ((n (temporary-filename x)))
      (funcall dumper arg n)
      n))
  (defun temporary-file-from-stream (i x)
    (temporary-file-from-foo #'dump-stream-to-file i x))
  (defun temporary-file-from-sexp (i x)
    (temporary-file-from-foo #'dump-sexp-to-file i x))
  (defun temporary-file-from-file (f x)
    (with-open-file (i f :direction :input :if-does-not-exist :error)
      (temporary-file-from-stream i x)))
  (defun ensure-lisp-file (x &optional (name "load.lisp"))
    (cond
      ((eq x t) (temporary-file-from-stream *standard-input* "load.lisp"))
      ((streamp x) (temporary-file-from-stream x "load.lisp"))
      ((eq x :self) (ensure-lisp-file-name *cl-launch-file* name))
      (t (ensure-lisp-file-name x name))))
  (defun ensure-lisp-file-name (x &optional (name "load.lisp"))
    (let ((p (pathname x)))
      (if (equal (pathname-type p) "lisp")
          p
          (temporary-file-from-file p name))))
  (defun cleanup-temporary-files ()
    (loop :for n = (pop *temporary-filenames*)
          :while n :do
          (ignore-errors (delete-file n)))))
(defun file-newer-p (new-file old-file)
  "Returns true if NEW-FILE is strictly newer than OLD-FILE."
  (> (file-write-date new-file) (file-write-date old-file)))
#-cormanlisp
(defun compile-and-load-file (source &key force-recompile
                              (verbose *verbose*) (load t)
                              output-file
                              #+(or gcl ecl) system-p)
  "compiles and load specified SOURCE file, if either required by keyword
argument FORCE-RECOMPILE, or not yet existing, or not up-to-date.
Keyword argument VERBOSE specifies whether to be verbose.
Returns two values: the fasl path, and T if the file was (re)compiled"

  ;; When in doubt, don't trust - recompile. Indeed, there are
  ;; edge cases cases when on the first time of compiling a simple
  ;; auto-generated file (e.g. from the automated test suite), the
  ;; fasl ends up being written to disk within the same second as the
  ;; source was produced, which cannot be distinguished from the
  ;; reverse case where the source code was produced in the same split
  ;; second as the previous version was done compiling. Could be
  ;; tricky if a big system needs be recompiled as a dependency on an
  ;; automatically generated file, but for cl-launch those
  ;; dependencies are not detected anyway (BAD). If/when they are, and
  ;; lacking better timestamps than the filesystem provides, you
  ;; should sleep after you generate your source code.
  ;;
  ;; Note: we don't seem to be using system-p for GCL, because it can dump core.
  #+(and gcl (not gcl-pre2.7))
  (setf source (ensure-lisp-file-name source (concatenate 'string (pathname-name source) ".lisp")))
  (let* ((truesource (truename source))
         (fasl
          (or output-file
              (compile-file-pathname*
               truesource
               #+(or gcl ecl) #+(or gcl ecl) :system-p system-p)))
         (compiled-p
          (when (or force-recompile
                    (not (probe-file fasl))
                    (not (file-newer-p fasl source)))
            (ensure-directories-exist fasl)
            (multiple-value-bind (path warnings failures)
                (compile-file truesource
                              :output-file fasl
                              #+(or gcl ecl) #+(or gcl ecl) :system-p system-p
                              #-gcl-pre2.7 #-gcl-pre2.7 #-gcl-pre2.7 #-gcl-pre2.7 :print verbose :verbose verbose)
              (declare (ignorable warnings failures))
              (unless (equal (truename fasl) (truename path))
                (error "CL-Launch: file compiled to ~A, expected ~A" path fasl))
              ;;#-lispworks
              (when failures
                (error "CL-Launch: failures while compiling ~A" source)))
            t)))
    (when load
      #+(or gcl ecl)
      (when system-p
        (return-from compile-and-load-file
          (values fasl (or compiled-p (nth-value 1 (compile-and-load-file source :force-recompile force-recompile :verbose verbose :load t))))))
      (load fasl :verbose verbose))
    (values fasl compiled-p)))
#+cormanlisp
(defun compile-and-load-file (source &key force-recompile verbose load output-file)
  "Corman Lisp has trouble with compiled files (says SLIME)."
  (declare (ignore force-recompile output-file))
  (when load
    (load source :verbose verbose))
  (force-output)
  (values nil t))
(defun load-file (source &key output-file)
  (declare (ignorable output-file))
  #-(or gcl-pre2.7 (and ecl (not dlopen)))
  (compile-and-load-file source :verbose *verbose* :output-file output-file)
  #+gcl-pre2.7
  (let* ((pn (parse-namestring source))) ; when compiling, gcl 2.6 will always
    (if (pathname-type pn) ; add a type .lsp if type is missing, to avoid compilation
      (compile-and-load-file source :verbose *verbose* :output-file output-file)
      (load source :verbose *verbose*)))
  #+(and ecl (not dlopen))
  (load source :verbose *verbose*))
(defparameter *asdf-attempts* '())
(defparameter *asdf-path* nil)
(defun asdf-call (x &rest args)
  (apply (find-symbol (string x) :asdf) args))
(defun recent-asdf-p ()
  (and (member :asdf2 *features*)
       (asdf-call :version-satisfies (asdf-call :asdf-version) "2.000")))
;;;; Load ASDF
(labels
    ((in-user-dir (x)
       (merge-pathnames x (user-homedir-pathname)))
     (try-asdf (thunk)
       (handler-bind (((or style-warning warning) 'muffle-warning))
         (funcall thunk))
       (recent-asdf-p))
     (try-implementation-asdf ()
       (push "Attempted to use the implementation-provided ASDF" *asdf-attempts*)
       (try-asdf (lambda () (ignore-errors (require :asdf) (setf *asdf-path* :required)))))
     (asdf-fasl (x)
       (compile-file-pathname
        (format nil "~A/~A--~(~A~)-~A.lisp"
                (or (getenv "TMP") "/tmp")
                (pathname-name x)
                (delete-if-not #'alphanumericp (lisp-implementation-type))
                (or (getenv "USERNAME") (getenv "USER")
                    (getenv "LOGNAME") (getenv "CL_LAUNCH_PID")))))
     (load-asdf-file (x)
       (load-file x :output-file (asdf-fasl x)))
     (try-asdf-file (x)
       (when (and x (probe-file x))
         (push (format nil "Attempted to load ASDF from ~A" x) *asdf-attempts*)
         (try-asdf (lambda ()
                     (and (load-asdf-file x)
                          #+ecl (load-asdf-file (make-pathname :name "asdf-ecl" :defaults x)))
                     (setf *asdf-path* x))))))
  #-gcl-pre2.7
  (block :a
    (when (or (recent-asdf-p) (try-implementation-asdf)) (return-from :a))
    (dolist (x (list ;;; TODO: follow XDG specification...
                (getenv "ASDF_PATH")
                (in-user-dir #p".local/share/common-lisp/source/asdf/asdf.lisp")
                (in-user-dir #p".local/share/common-lisp/source/cl-asdf/asdf.lisp")
                (in-user-dir #p"cl/asdf/asdf.lisp")
                #p"/usr/share/common-lisp/source/cl-asdf/asdf.lisp"
                #p"/usr/share/common-lisp/source/asdf/asdf.lisp"))
      (when (try-asdf-file x) (return-from :a)))))
;; Even in absence of a recent ASDF, at least create a package asdf.
(unless (find-package :asdf)
  (make-package :asdf)))
;;;; CL-Launch Initialization code
(progn

;;#+ecl (require 'cmp) ; ensure we use the compiler (we use e.g. *ecl-library-directory*)

;; To dynamically recompute from the command-line at each invocation
(defvar *arguments* nil "command-line parameters")

;; Variables that define the current system
(defvar *dumped* nil)
(defvar *restart* nil)
(defvar *init-forms* nil)
(defvar *quit* t)

(defun raw-command-line-arguments ()
  #+ecl (loop :for i :from 0 :below (si:argc) :collect (si:argv i))
  #+gcl si:*command-args*
  #+cmu extensions:*command-line-strings*
  #+clozure (ccl::command-line-arguments)
  #+sbcl sb-ext:*posix-argv*
  #+allegro (sys:command-line-arguments) ; default: :application t
  #+lispworks sys:*line-arguments-list*
  #+clisp (coerce (ext:argv) 'list)
  #-(or ecl gcl cmu clozure sbcl allegro lispworks clisp)
  (error "raw-command-line-arguments not implemented yet"))

(defun command-line-arguments ()
  (let* ((raw (raw-command-line-arguments))
         (cooked
          #+(or sbcl allegro) raw
          #-(or sbcl allegro)
          (if (eq *dumped* :standalone)
              raw
              (member "--" raw :test 'string-equal))))
    (cdr cooked)))

(defun compute-arguments ()
  (setf *cl-launch-file* (getenv "CL_LAUNCH_FILE")
        *temporary-directory* (ensure-directory-name (or (getenv "TMP") "/tmp"))
        #+gcl #+gcl system::*tmp-dir* *temporary-directory* ; basic lack fixed after gcl 2.7.0-61, but ending / required still on 2.7.0-64.1
        *verbose* (when (getenv "CL_LAUNCH_VERBOSE") t)
        *arguments* (or *arguments* (command-line-arguments))))

(defun load-stream (&optional (s #-clisp *standard-input*
				 #+clisp *terminal-io*))
  ;; GCL 2.6 can't load from a string-input-stream
  ;; ClozureCL 1.2 cannot load from either *standard-input* or *terminal-io*
  ;; Allegro 5, I don't remember but it must have been broken when I tested.
  #+(or gcl-pre2.7 clozure allegro)
  (do ((eof '#:eof) (x t (read s nil eof))) ((eq x eof)) (eval x))
  #-(or gcl-pre2.7 clozure allegro)
  (load s :verbose *verbose* :print *verbose*))

(defun load-string (string)
  (with-input-from-string (s string) (load-stream s)))

(defun finish-outputs ()
  (finish-output *error-output*)
  (finish-output))

(defun %abort (code fmt &rest args)
  (apply #'format *error-output* fmt args)
  (quit code))

(defun resume (&rest keys &key restart init-forms quit)
  (declare (ignore restart init-forms quit))
  (compute-arguments)
  (apply #'do-resume keys))

(defun do-resume (&key (restart *restart*) (init-forms *init-forms*) (quit *quit*))
  (when restart (funcall restart))
  (when init-forms (load-string init-forms))
  (finish-outputs)
  (when quit (quit 0)))

(defun dump-image (filename &key standalone (package :cl-user))
  (setf *dumped* (if standalone :standalone :wrapped)
        *arguments* nil
	*package* (find-package package)
	*features* (remove :cl-launched *features*))
  #+clisp
  (apply #'ext:saveinitmem filename
   :quiet t
   :start-package *package*
   :keep-global-handlers nil
   :executable (if standalone 0 t) ;--- requires clisp 2.48 or later.
   (when standalone
     (list
      :norc t
      :script nil
      :init-function #'resume
      ;; :parse-options nil ;--- requires a non-standard patch to clisp.
      )))
  #+sbcl
  (progn
    ;;(sb-pcl::precompile-random-code-segments) ;--- it is ugly slow at compile-time (!) when the initial core is a big CLOS program. If you want it, do it yourself
   (setf sb-ext::*gc-run-time* 0)
   (apply 'sb-ext:save-lisp-and-die filename
    :executable t ;--- always include the runtime that goes with the core
    (when standalone (list :toplevel #'resume :save-runtime-options t)))) ;--- only save runtime-options for standalone executables
  #+cmu
  (progn
   (ext:gc :full t)
   (setf ext:*batch-mode* nil)
   (setf ext::*gc-run-time* 0)
   (extensions:save-lisp filename))
  #+clozure
  (ccl:save-application filename :prepend-kernel t)
  #+allegro
  (progn
   (sys:resize-areas :global-gc t :pack-heap t :sift-old-areas t :tenure t) ; :new 5000000
   (excl:dumplisp :name filename :suppress-allegro-cl-banner t))
  #+lispworks
  (progn
    (system::copy-file (getenv "LWLICENSE")
                       (make-pathname :name "lwlicense" :type nil :defaults filename))
    (if standalone
        (lispworks:deliver 'resume filename 0 :interface nil)
        (hcl:save-image filename :environment nil)))
  #+gcl
  (progn
   (si::set-hole-size 500) (si::gbc nil) (si::sgc-on t)
   (si::save-system filename))
  #-(or clisp sbcl cmu clozure allegro gcl lispworks)
  (%abort 11 "Can't dump ~S: cl-launch doesn't supports image dumping with this Lisp implementation.~%" filename))

(defun run (&key source-registry load system dump restart init (quit 0))
  (pushnew :cl-launched *features*)
  (compute-arguments)
  (when (and (recent-asdf-p) (or source-registry system))
    (asdf-call :initialize-source-registry source-registry))
  (if dump
      (build-and-dump dump load system restart init quit)
      (build-and-run load system restart init quit)))

(defun read-function (string)
  (eval `(function ,(read-from-string string))))

#-(and gcl (not gcl-pre2.7))
(defun build-and-load (load system restart init quit)
  (do-build-and-load load system restart init quit))

#+(and gcl (not gcl-pre2.7))
(defun build-and-load (load system restart init quit)
  (unwind-protect
       (do-build-and-load load system restart init quit)
    (cleanup-temporary-files)))

(defun do-build-and-load (load system restart init quit)
  (etypecase load
    (null nil)
    ((eql t) (load-stream))
    (stream (load-stream load))
    ((eql :self) (load-file *cl-launch-file*))
    ((or pathname string) (load-file load)))
  (when system
    (load-systems system))
  (setf *restart* (when restart (read-function restart))
        *init-forms* init
        *quit* quit))

(defun build-and-run (load system restart init quit)
  (build-and-load load system restart init quit)
  (do-resume))

#-ecl
(defun build-and-dump (dump load system restart init quit)
  (build-and-load load system restart init quit)
  (dump-image dump :standalone (getenv "CL_LAUNCH_STANDALONE"))
  (quit))

(defun ensure-directory-name (dn)
  (if (eql #\/ (char dn (1- (length dn)))) dn
      (concatenate 'string dn "/")))

;;; Attempt at compiling directly with asdf-ecl's make-build and temporary wrapper asd's
#+ecl (defvar *in-compile* nil)
#+ecl
(defun build-and-dump (dump load system restart init quit)
  (unwind-protect
       (let* ((*compile-verbose* *verbose*)
              (*in-compile* t)
              (c::*suppress-compiler-warnings* (not *verbose*))
              (c::*suppress-compiler-notes* (not *verbose*))
              (*features* (remove :cl-launch *features*))
              (header (or *compile-file-pathname* *load-pathname* (getenv "CL_LAUNCH_HEADER")))
              (header-file (ensure-lisp-file header "header.lisp"))
              (load-file (when load (ensure-lisp-file load "load.lisp")))
              (standalone (getenv "CL_LAUNCH_STANDALONE"))
              (init-code
               `(unless *in-compile*
                 (setf
                  *package* (find-package :cl-user)
                  *load-verbose* nil
                  *dumped* ,(if standalone :standalone :wrapped)
                  *arguments* nil
                  asdf::*source-registry* nil asdf::*output-translations* nil
                  ,@(when restart `(*restart* (read-function ,restart)))
                  ,@(when init `(*init-forms* ,init))
                  ,@(unless quit `(*quit* nil)))
                 ,(if standalone '(resume) '(si::top-level))))
              (init-file (temporary-file-from-sexp init-code "init.lisp"))
              (prefix-sys (pathname-name (temporary-filename "prefix")))
              (program-sys (pathname-name (temporary-filename "program")))
              (prefix-sysdef
               `(asdf::defsystem ,prefix-sys
                 :depends-on () :serial t
                 :components ((:file "header" :pathname ,(truename header-file))
                              ,@(when load-file `((:file "load" :pathname ,(truename load-file)))))))
              (program-sysdef
               `(asdf::defsystem ,program-sys
                 :depends-on (,prefix-sys
                              ,@(when system `(,system))
                              ,prefix-sys) ;; have the prefix first, whichever order asdf traverses
                 :components ((:file "init" :pathname ,(truename init-file)))))
              (prefix-asd (temporary-file-from-sexp prefix-sysdef "prefix.asd"))
              (program-asd (temporary-file-from-sexp program-sysdef "program.asd")))
         (load prefix-asd)
         (load program-asd)
         (asdf::make-build program-sys :type :program)
         (si:system (format nil "cp -p ~S ~S"
                            (namestring (first (asdf::output-files (make-instance 'asdf::program-op)
                                                                  (asdf::find-system program-sys))))
                            dump)))
    (cleanup-temporary-files))
  (quit))

#|
;; We provide cl-launch, no need to go looking for it further!
(when (recent-asdf-p)
  (unless (asdf::find-system :cl-launch nil)
    (eval `(asdf::defsystem :cl-launch
             :pathname ,(or *compile-file-truename* *load-truename*
                            #-windows "/dev/null"
                            #+windows "\\NUL")
             :depends-on () :serial t :components ()))))
|#

(defun load-systems (&rest systems)
  (if (recent-asdf-p)
      (dolist (s systems) (asdf-call :load-system s :verbose *verbose*))
      (%abort 10 "ERROR: ASDF requested, but ASDF 2 not found~%~{~A~%~}"
              (reverse *asdf-attempts*))))


;;#+ecl (progn (trace c::builder c::build-fasl c:build-static-library c:build-program ensure-lisp-file-name ensure-lisp-file cleanup-temporary-files asdf:operate asdf:perform asdf::bundle-sub-operations asdf::gather-components asdf::traverse asdf:output-files asdf:input-files) (setf c::*compiler-break-enable* t))
;;(progn (trace make-package load-file load-stream load-systems build-and-dump build-and-run run resume compute-arguments do-resume compile-file compile-and-load-file compile-file-pathname* load asdf:operate asdf:perform) (setf *verbose* t *load-verbose* t *compile-verbose* t))

(when (recent-asdf-p)
  (setf (symbol-function 'compile-file-pathname*) (function asdf::compile-file-pathname*)))
(pushnew :cl-launch *features*))

(cl-launch::compute-arguments)

;;; END OF CL-LAUNCH LISP HEADER
